import os
from hashlib import sha256
from pathlib import Path

from cryptography.hazmat.primitives import serialization
from cryptography.x509.base import load_pem_x509_certificate
from pytest import raises

from pyattest.attestation import Attestation
from pyattest.configs.apple import AppleConfig
from pyattest.exceptions import InvalidAaguidException, InvalidNonceException, InvalidKeyIdException, \
    InvalidAppIdException, InvalidCounterException, InvalidCredentialIdException
from pyattest.testutils.factories.attestation import apple as apple_factory

root_ca = load_pem_x509_certificate(Path('pyattest/testutils/fixtures/root_cert.pem').read_bytes())
root_ca_pem = root_ca.public_bytes(serialization.Encoding.PEM)
nonce = os.urandom(32)


def test_happy_path():
    """ Test the basic attest verification where everything works like it should :) """
    attest, public_key = apple_factory.get(app_id='foo', nonce=nonce)
    key_id = sha256(public_key).digest()
    config = AppleConfig(key_id=key_id, app_id='foo', root_ca=root_ca_pem)

    attestation = Attestation(attest, nonce, config)
    attestation.verify()


def test_aaguid():
    attest, public_key = apple_factory.get(app_id='foo', nonce=nonce)
    key_id = sha256(public_key).digest()
    config = AppleConfig(key_id=key_id, app_id='foo', root_ca=root_ca_pem,
                         production=True)

    attestation = Attestation(attest, nonce, config)
    with raises(InvalidAaguidException):
        attestation.verify()


def test_nonce():
    attest, public_key = apple_factory.get(app_id='foo', nonce=nonce)
    key_id = sha256(public_key).digest()
    config = AppleConfig(key_id=key_id, app_id='foo', root_ca=root_ca_pem)

    attestation = Attestation(attest, os.urandom(32), config)
    with raises(InvalidNonceException):
        attestation.verify()


def test_key_id():
    attest, public_key = apple_factory.get(app_id='foo', nonce=nonce)
    key_id = sha256(b'invalid_public_key').digest()
    config = AppleConfig(key_id=key_id, app_id='foo', root_ca=root_ca_pem)

    attestation = Attestation(attest, nonce, config)
    with raises(InvalidKeyIdException):
        attestation.verify()


def test_app_id():
    attest, public_key = apple_factory.get(app_id='foo', nonce=nonce)
    key_id = sha256(public_key).digest()
    config = AppleConfig(key_id=key_id, app_id='bar', root_ca=root_ca_pem)

    attestation = Attestation(attest, nonce, config)
    with raises(InvalidAppIdException):
        attestation.verify()


def test_counter():
    attest, public_key = apple_factory.get(app_id='foo', nonce=nonce, aaguid=b'appattestdevelop', counter=9)
    key_id = sha256(public_key).digest()
    config = AppleConfig(key_id=key_id, app_id='foo', root_ca=root_ca_pem)

    attestation = Attestation(attest, nonce, config)
    with raises(InvalidCounterException):
        attestation.verify()


def test_credential_id():
    attest, public_key = apple_factory.get(app_id='foo', nonce=nonce, wrong_public_key=True)
    key_id = sha256(public_key).digest()
    config = AppleConfig(key_id=key_id, app_id='foo', root_ca=root_ca_pem)

    attestation = Attestation(attest, nonce, config)
    with raises(InvalidCredentialIdException):
        attestation.verify()



import base64
from pathlib import Path

from cryptography.hazmat.primitives import serialization
from cryptography.x509.base import load_pem_x509_certificate
from pytest import raises

from pyattest.attestation import Attestation
from pyattest.configs.google import GoogleConfig
from pyattest.exceptions import InvalidKeyIdException, InvalidCtsProfile, InvalidBasicIntegrity, \
    InvalidCertificateChainException, InvalidAppIdException
import os

from pyattest.testutils.factories.attestation import google as google_factory

root_cn = 'pyattest-testing-leaf.ch'
root_ca = load_pem_x509_certificate(Path('pyattest/testutils/fixtures/root_cert.pem').read_bytes())
root_ca_pem = root_ca.public_bytes(serialization.Encoding.PEM)
nonce = os.urandom(32)


def test_happy_path():
    """ Test the basic attest verification where everything works like it should :) """
    attest, key_id = google_factory.get(apk_package_name='foo', nonce=nonce)
    config = GoogleConfig(key_ids=[base64.b64encode(key_id)], apk_package_name='foo', root_cn=root_cn,
                          root_ca=root_ca_pem, production=False)

    attestation = Attestation(attest, nonce, config)
    attestation.verify()


def test_key_id():
    attest, key_id = google_factory.get(apk_package_name='foo', nonce=nonce)
    config = GoogleConfig(key_ids=[base64.b64encode(b'100%wrong')], apk_package_name='foo', root_cn=root_cn,
                          root_ca=root_ca_pem, production=False)

    attestation = Attestation(attest, nonce, config)

    with raises(InvalidKeyIdException):
        attestation.verify()


def test_cts_profile():
    attest, key_id = google_factory.get(apk_package_name='foo', nonce=nonce, cts_profile=False)
    config = GoogleConfig(key_ids=[base64.b64encode(key_id)], apk_package_name='foo', root_cn=root_cn,
                          root_ca=root_ca_pem, production=True)

    attestation = Attestation(attest, nonce, config)

    with raises(InvalidCtsProfile):
        attestation.verify()


def test_basic_integrity():
    attest, key_id = google_factory.get(apk_package_name='foo', nonce=nonce, basic_integrity=False)
    config = GoogleConfig(key_ids=[base64.b64encode(key_id)], apk_package_name='foo', root_cn=root_cn,
                          root_ca=root_ca_pem, production=True)

    attestation = Attestation(attest, nonce, config)

    with raises(InvalidBasicIntegrity):
        attestation.verify()


def test_apk_package_name():
    attest, key_id = google_factory.get(apk_package_name='foo', nonce=nonce)
    config = GoogleConfig(key_ids=[base64.b64encode(key_id)], apk_package_name='bar', root_cn=root_cn,
                          root_ca=root_ca_pem, production=True)

    attestation = Attestation(attest, nonce, config)

    with raises(InvalidAppIdException):
        attestation.verify()


def test_certificate_chain():
    attest, key_id = google_factory.get(apk_package_name='foo', nonce=nonce)
    config = GoogleConfig(key_ids=[base64.b64encode(key_id)], apk_package_name='bar', root_cn=root_cn, production=True)

    attestation = Attestation(attest, nonce, config)

    with raises(InvalidCertificateChainException):
        attestation.verify()



import base64
import os

from _pytest.python_api import raises
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

from pyattest.attestation import Attestation
from pyattest.configs.google_play_integrity_api import GooglePlayIntegrityApiConfig
from pyattest.exceptions import PyAttestException, InvalidNonceException, InvalidAppIdException, InvalidAppIntegrity, \
    InvalidDeviceIntegrity
from pyattest.testutils.factories.attestation.google import get_play_integrity_api_attestation

nonce = os.urandom(32)


def test_happy_path():
    apk_package_name = 'foo'
    attestation, verification_key, aes_key = get_play_integrity_api_attestation(apk_package_name, nonce,
                                                                                'PLAY_RECOGNIZED',
                                                                                ['MEETS_DEVICE_INTEGRITY'])
    config = GooglePlayIntegrityApiConfig(decryption_key=base64.standard_b64encode(aes_key).decode(),
                                          verification_key=base64.standard_b64encode(
                                              verification_key.public_bytes(encoding=Encoding.DER,
                                                                            format=PublicFormat.SubjectPublicKeyInfo))
                                          .decode(),
                                          apk_package_name=apk_package_name, production=True)

    attestation = Attestation(attestation, nonce, config)
    attestation.verify()


def test_invalid_signed_attest_token():
    apk_package_name = 'foo'
    attestation, verification_key, aes_key = get_play_integrity_api_attestation(apk_package_name, nonce,
                                                                                'PLAY_RECOGNIZED',
                                                                                ['MEETS_DEVICE_INTEGRITY'])
    wrong_aes_key = os.urandom(32)
    config = GooglePlayIntegrityApiConfig(decryption_key=base64.standard_b64encode(wrong_aes_key).decode(),
                                          verification_key=base64.standard_b64encode(
                                              verification_key.public_bytes(encoding=Encoding.DER,
                                                                            format=PublicFormat.SubjectPublicKeyInfo))
                                          .decode(),
                                          apk_package_name=apk_package_name, production=True)

    attestation = Attestation(attestation, nonce, config)
    with raises(PyAttestException):
        attestation.verify()


def test_invalid_nonce():
    apk_package_name = 'foo'
    invalid_nonce = os.urandom(32)
    attestation, verification_key, aes_key = get_play_integrity_api_attestation(apk_package_name, invalid_nonce,
                                                                                'PLAY_RECOGNIZED',
                                                                                ['MEETS_DEVICE_INTEGRITY'])
    config = GooglePlayIntegrityApiConfig(decryption_key=base64.standard_b64encode(aes_key).decode(),
                                          verification_key=base64.standard_b64encode(
                                              verification_key.public_bytes(encoding=Encoding.DER,
                                                                            format=PublicFormat.SubjectPublicKeyInfo))
                                          .decode(),
                                          apk_package_name=apk_package_name, production=True)
    attestation = Attestation(attestation, nonce, config)
    with raises(InvalidNonceException):
        attestation.verify()


def test_invalid_apk_package():
    apk_package_name = 'foo'
    wrong_apk_package_name = 'foo_'
    attestation, verification_key, aes_key = get_play_integrity_api_attestation(wrong_apk_package_name, nonce,
                                                                                'PLAY_RECOGNIZED',
                                                                                ['MEETS_DEVICE_INTEGRITY'])
    config = GooglePlayIntegrityApiConfig(decryption_key=base64.standard_b64encode(aes_key).decode(),
                                          verification_key=base64.standard_b64encode(
                                              verification_key.public_bytes(encoding=Encoding.DER,
                                                                            format=PublicFormat.SubjectPublicKeyInfo))
                                          .decode(),
                                          apk_package_name=apk_package_name, production=True)
    attestation = Attestation(attestation, nonce, config)
    with raises(InvalidAppIdException):
        attestation.verify()


def test_invalid_app_integrity():
    apk_package_name = 'foo'
    attestation, verification_key, aes_key = get_play_integrity_api_attestation(apk_package_name, nonce,
                                                                                'PLAY_UNRECOGNIZED',
                                                                                ['MEETS_DEVICE_INTEGRITY'])
    config = GooglePlayIntegrityApiConfig(decryption_key=base64.standard_b64encode(aes_key).decode(),
                                          verification_key=base64.standard_b64encode(
                                              verification_key.public_bytes(encoding=Encoding.DER,
                                                                            format=PublicFormat.SubjectPublicKeyInfo))
                                          .decode(),
                                          apk_package_name=apk_package_name, production=True)
    attestation = Attestation(attestation, nonce, config)
    with raises(InvalidAppIntegrity):
        attestation.verify()


def test_invalid_device_integrity():
    apk_package_name = 'foo'
    attestation, verification_key, aes_key = get_play_integrity_api_attestation(apk_package_name, nonce,
                                                                                'PLAY_RECOGNIZED',
                                                                                [])
    config = GooglePlayIntegrityApiConfig(decryption_key=base64.standard_b64encode(aes_key).decode(),
                                          verification_key=base64.standard_b64encode(
                                              verification_key.public_bytes(encoding=Encoding.DER,
                                                                            format=PublicFormat.SubjectPublicKeyInfo))
                                          .decode(),
                                          apk_package_name=apk_package_name, production=True)
    attestation = Attestation(attestation, nonce, config)
    with raises(InvalidDeviceIntegrity):
        attestation.verify()



import base64
import json
from typing import Optional

from jose import jwe, jws

from pyattest.exceptions import PyAttestException, InvalidNonceException, \
    InvalidAppIdException, InvalidAppIntegrity, \
    InvalidDeviceIntegrity
from pyattest.verifiers.attestation import AttestationVerifier


class GooglePlayIntegrityAttestationVerifier(AttestationVerifier):
    def verify(self):
        """
        Verify the given attestation based on the Google documentation. The attestation is a nested JWS in JWE object.

        Raises a PyAttestException as soon as one of the verification steps fails.

        See also:
        Google Documentation https://developer.android.com/google/play/integrity/verdict#decrypt-verify
        """
        payload = self.unpack(self.attestation.raw)
        self.check_request_details(payload.get('requestDetails'))
        if self.attestation.config.production:
            self.check_app_integrity(payload.get('appIntegrity'))
            self.check_device_integrity(payload.get('deviceIntegrity'))

    def unpack(self, jwt_object: str) -> dict:
        try:
            jwe_token = jwt_object
            jws_token = jwe.decrypt(jwe_token, self.attestation.config.decryption_key)
            payload = jws.verify(jws_token, self.attestation.config.verification_key, 'ES256')
            return json.loads(payload)
        except Exception as exception:
            raise PyAttestException from exception

    def check_app_integrity(self, app_integrity):
        if app_integrity.get('appRecognitionVerdict') != 'PLAY_RECOGNIZED':
            raise InvalidAppIntegrity
        self.verify_apk_package_name(app_integrity.get('packageName'))

    def check_device_integrity(self, device_integrity):
        if 'MEETS_DEVICE_INTEGRITY' not in device_integrity.get('deviceRecognitionVerdict'):
            raise InvalidDeviceIntegrity

    def check_request_details(self, request_details):
        self.verify_nonce(request_details.get('nonce'))
        self.verify_apk_package_name(request_details.get('requestPackageName'))

    def verify_nonce(self, nonce: str):
        if base64.urlsafe_b64decode(nonce) != self.attestation.nonce:
            raise InvalidNonceException

    def verify_apk_package_name(self, package_name: Optional[str]):
        if not package_name or package_name != self.attestation.config.apk_package_name:
            raise InvalidAppIdException



import base64

from cryptography.hazmat.primitives.serialization import load_der_public_key

from pyattest.configs.config import Config
from pyattest.verifiers.google_assertion import GoogleAssertionVerifier
from pyattest.verifiers.google_play_integrity_attestation import GooglePlayIntegrityAttestationVerifier


class GooglePlayIntegrityApiConfig(Config):
    attestation_verifier_class = GooglePlayIntegrityAttestationVerifier
    assertion_verifier_class = GoogleAssertionVerifier

    def __init__(self, decryption_key: str, verification_key: str, apk_package_name: str, production: bool = False):
        self.decryption_key = base64.standard_b64decode(decryption_key)
        self.verification_key = load_der_public_key(base64.standard_b64decode(verification_key))
        self.apk_package_name = apk_package_name
        self.production = production
